/* tcp-webserver-example.cc
// TE816 - Modelagem e Avaliação de Desempenho
// Pós Graduação em Engenharia Elétrica
// Universidade Federal do Paraná
// This souce was developed as basis for homework 
//
// -- 10/11/2015
// ==================================
// Prof. Carlos Marcelo Pedroso
// pedroso@eletrica.ufpr.br
// http://www.eletrica.ufpr.br/pedroso
// Universidade Federal do Paraná
// Departamento de Engenharia Elétrica
// Centro Politécnico
// CEP: 81531-990 - Curitiba - PR - Brasil
 */

#include <iostream>
#include <fstream>

#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/internet-module.h"
#include "ns3/point-to-point-module.h"
#include "ns3/applications-module.h"
#include "ns3/flow-monitor-helper.h"

using namespace ns3;

NS_LOG_COMPONENT_DEFINE ("TcpWebExample");

// ===========================================================================
//
//         node 0                 node 1
//   +----------------+    +----------------+
//   |    ns-3 TCP    |    |    ns-3 TCP    |
//   +----------------+    +----------------+
//   |    10.1.1.1    |    |    10.1.1.2    |
//   +----------------+    +----------------+
//   | point-to-point |    | point-to-point |
//   +----------------+    +----------------+
//           |                     |
//           +---------------------+
//                100 Mbps, 1 ms
//
// This is a simple test of TCP connection. We create two nodes,
// one will be a TCP client and other the TCP server, 
// The nodes transmit data in the simple point-to-point channel. 
//
// ===========================================================================
//

int
main (int argc, char *argv[])
{
  
  //
  // The three lines below enable debugging mode. Comment these three lines for disable.
  //
  LogComponentEnable ("TcpWebExample", LOG_LEVEL_INFO);
  LogComponentEnable ("TcpWebServerApplication", LOG_LEVEL_INFO);
  LogComponentEnable ("TcpWebClientApplication", LOG_LEVEL_INFO);

  //
  // Allow the user to override any of the defaults and the above Bind() at
  // run-time, via command-line arguments
  //
  bool useV6 = false;
  Address serverAddress;

  CommandLine cmd;
  cmd.AddValue ("useIpv6", "Use Ipv6", useV6);
  cmd.Parse (argc, argv);
  
  //
  // Create two nodes required by the topology (point-to-point).
  //
  NS_LOG_INFO ("Create nodes.");
  NodeContainer nodes;
  nodes.Create (2);

  //
  // Create and configure channel for the communication.
  //
  NS_LOG_INFO("Create and configuring point to point link.");
  PointToPointHelper p2p;
  p2p.SetDeviceAttribute ("DataRate", StringValue ("100Mbps"));
  p2p.SetChannelAttribute ("Delay", StringValue ("1ms"));
  p2p.SetDeviceAttribute ("Mtu", UintegerValue (1500));

  NetDeviceContainer devices;
  devices = p2p.Install (nodes);

  InternetStackHelper stack;
  stack.Install (nodes);
  
  //
  // Now, add IP address in the nodes.
  //
  NS_LOG_INFO("Assign IP Address.");
  if (useV6 == false)
    {
      Ipv4AddressHelper ipv4;
      ipv4.SetBase ("10.1.1.0", "255.255.255.0");
      Ipv4InterfaceContainer i = ipv4.Assign (devices);
      serverAddress = Address(i.GetAddress (1));
    }
  else
    {
      Ipv6AddressHelper ipv6;
      ipv6.SetBase ("2001:0000:f00d:cafe::", 64);
      Ipv6InterfaceContainer i6 = ipv6.Assign (devices);
      serverAddress = Address(i6.GetAddress (1,1));
    }
  
  //
  // Create a TcpWebClientHelperServer on node 1.
  //
  NS_LOG_INFO("Setting Server Application.");
  uint16_t port = 80; // well-known web port number.
  TcpWebServerHelper webServerHelper (port);
  webServerHelper.SetAttribute ("PageSize", UintegerValue (5000));
  ApplicationContainer serverApps = webServerHelper.Install (nodes.Get (1));
  serverApps.Start (Seconds (0.0));
  serverApps.Stop (Seconds (10.0));
  
  // Create a webClientHelper application
  NS_LOG_INFO("Setting Client 1.");
  TcpWebClientHelper webClientHelper1 (serverAddress, port);
  webClientHelper1.SetAttribute ("MaxRequests", UintegerValue (5));
  webClientHelper1.SetAttribute ("Interval", TimeValue (Seconds (1.0)));
  webClientHelper1.SetAttribute ("RequestSize", UintegerValue (200));
  ApplicationContainer clientApps1 = webClientHelper1.Install (nodes.Get (0));
  clientApps1.Start (Seconds (1.0));
  clientApps1.Stop (Seconds (10.0));

  // Create a webClientHelper application
  //NS_LOG_INFO("Setting Client 2.");
  //TcpWebClientHelper webClientHelper2 (serverAddress, port);
  //webClientHelper2.SetAttribute ("MaxRequests", UintegerValue (5));
  //webClientHelper2.SetAttribute ("Interval", TimeValue (Seconds (1.0)));
  //webClientHelper2.SetAttribute ("RequestSize", UintegerValue (200));
  //ApplicationContainer clientApps2 = webClientHelper2.Install (nodes.Get (0));
  //clientApps2.Start (Seconds (0.5));
  //clientApps2.Stop (Seconds (10.0));

  //
  // Uncomment to enable packet trace in pcap format and save on file TcpWebClientHelper_example.pcap.
  // Comment the two lines below to disable this.
  // AsciiTraceHelper ascii;
   p2p.EnablePcapAll("TcpWebClientHelper_example");

  // Flow monitor
  FlowMonitorHelper flowHelper;
  flowHelper.InstallAll ();
    
  Simulator::Stop (Seconds (11.0)); 

  // Start the simulation.
  NS_LOG_INFO("Start Simulation.");
  Simulator::Run ();

  flowHelper.SerializeToXmlFile ("TcpWebServer.flowmonitor", true, true);

  Simulator::Destroy ();
  NS_LOG_INFO("Simulation finished.");
  return 0;
}
